import re
from os.path import dirname, exists, join
from random import choice as rchoice
from random import randint as rint
from random import randrange as rrange
from typing import Dict, List

import httpx
import orjson

'''
Fake header generator

Usage:
>>> from playfulrequests import Headers
>>> h = Headers(os='win', browser='chrome')
>>> h.generate()
'''


class OSHeaders:
    @staticmethod
    def windows() -> str:
        version = ('10.0', f'6.{rrange(4)}')[rrange(2)]

        if version == '10.0' or rrange(2):
            version += f"; {('WOW64', 'Win64; x64')[rrange(2)]}"

        return f'Windows NT {version}'

    @staticmethod
    def macos() -> str:
        sub = str(rint(10, 14))
        sub += '_' + str(rint(1, (6 if sub != '14' else 2)))

        return f'Macintosh; Intel Mac OS X 10_{sub}'

    @staticmethod
    def linux() -> str:
        return 'X11; Linux ' + ('x86_64', 'i686', 'i686 on x86_64')[rrange(3)]

    @staticmethod
    def random_os() -> str:
        return (OSHeaders.windows, OSHeaders.macos, OSHeaders.linux)[rrange(3)]()


class VersionScraper:
    threshold: int = 10
    data: List[str]

    @staticmethod
    def leading_num(line: str) -> int:
        return int(line.split('.', 1)[0])

    def __init__(self) -> None:
        if not exists(self.file_name):
            self.data = self.download()
        self.data = self.load()

    def load(self) -> List[str]:
        with open(self.file_name, 'rb') as f:
            return orjson.loads(f.read())

    def write_file(self, data: List[str]) -> None:
        with open(self.file_name, 'wb') as f:
            f.write(orjson.dumps(data))


class ChromeVersions(VersionScraper):
    resource: str = 'https://raw.githubusercontent.com/vikyd/chromium-history-version-position/master/json/all-version.json'
    file_name: str = join(dirname(__file__), "bin", "CR_VERSIONS.json")

    @staticmethod
    def get_ver(line: str) -> str:
        return re.sub('[^\d\.]', '', line)

    def download(self) -> List[str]:
        print('Downloading Chrome version history...')
        versions: List[str]
        with httpx.stream('GET', self.resource) as r:
            lines = r.iter_lines()
            # search for first version number
            for line in lines:
                highest_ver = re.search('\d+', line)
                if highest_ver:  # on the first number
                    # get leading ver number
                    highest_ver = int(highest_ver[0])
                    break
            # add first item to the versions list
            versions = [self.get_ver(line)]
            # for each line in the stream
            # if the leading number is within 20 of the first number
            # add it to the versions list
            while self.leading_num(versions[-1]) >= highest_ver - self.threshold:
                versions.append(self.get_ver(next(lines)))
        # write the versions list to a file
        self.write_file(versions)
        return versions

    def generate(self) -> str:
        return (
            'Mozilla/5.0 (%PLAT%) AppleWebKit/537.36 (KHTML,'
            + f' like Gecko) Chrome/{rchoice(self.data)} Safari/537.36'
        )


class FirefoxVersions(VersionScraper):
    resource: str = 'https://ftp.mozilla.org/pub/firefox/releases/'
    file_name: str = join(dirname(__file__), "bin", "FF_VERSIONS.json")

    def download(self) -> List[str]:
        print('Downloading Firefox version history...')
        resp = httpx.get(self.resource)
        # gather version numbers
        versions: List[str]
        versions = re.findall(r'/pub/firefox/releases/([\d\.]+)/', resp.text)
        versions = sorted(set(versions), key=self.leading_num, reverse=True)

        highest_ver: int = self.leading_num(versions[0])
        # remove versions that have a major number
        # less than the highest version - threshold
        versions = [
            ver for ver in versions if self.leading_num(ver) >= highest_ver - self.threshold
        ]
        self.write_file(versions)
        return versions

    def generate(self) -> str:
        ver: str = rchoice(self.data)
        return f'Mozilla/5.0 (%PLAT%; rv:{ver}) Gecko/20100101 Firefox/{ver}'


chrome = ChromeVersions().generate
firefox = FirefoxVersions().generate


class Headers:
    '''
    browser - str, chrome/firefox. User Agent browser. Default: random
    os - str, win/mac/lin. OS of User Agent. Default: random
    headers - bool, True/False. Generate random headers or no. Default: False
    '''

    _os: dict = {'win': OSHeaders.windows, 'mac': OSHeaders.macos, 'lin': OSHeaders.linux}

    _browser: dict = {'chrome': chrome, 'firefox': firefox}

    def __init__(self, browser: str = None, os: str = None, headers: bool = True) -> None:
        self._platform: str = self._os.get(os, OSHeaders.random_os)
        self._browser: str = self._browser.get(browser, (chrome, firefox)[rrange(2)])
        self._headers: bool = headers

    @staticmethod
    def make_header() -> Dict[str, str]:
        return {key: value for key, value in header_template.items() if rrange(2)}

    def generate(self) -> dict:
        platform = self._platform()
        browser = self._browser()

        headers = {
            'Accept': '*/*',
            'Connection': 'keep-alive',
            'User-Agent': browser.replace('%PLAT%', platform),
        }

        self._headers and headers.update(self.make_header())

        return headers


header_template: Dict[str, str] = {
    'Accept-Encoding': 'gzip, deflate, br',
    'Accept-Language': 'en-US;q=0.5,en;q=0.3',
    'Cache-Control': 'max-age=0',
    'DNT': '1',
    'Upgrade-Insecure-Requests': '1',
    'Referer': 'https://google.com',
    'Pragma': 'no-cache',
}



import httpx
from async_class import AsyncObject, link


class SplitError(Exception):
    pass


class ProxyCheckError(Exception):
    pass


class ProxyManager(AsyncObject):
    async def __ainit__(self, inst, proxy) -> None:
        link(self, inst)

        self.proxy = proxy.strip() if proxy else None
        self.http_proxy = None
        self.ip = None
        self.port = None
        self.username = None
        self.password = None
        self.browser_proxy = None
        self.plain_proxy = None
        self.timeout = httpx.Timeout(20.0, read=None)

        if self.proxy:
            self.split_proxy()
            self.proxy = (
                f"{self.username}:{self.password}@{self.ip}:{self.port}"
                if self.username
                else f"{self.ip}:{self.port}"
            )
            self.plain_proxy = f"http://{self.proxy}"

            if self.username:
                self.browser_proxy = {
                    "server": self.plain_proxy,
                    "username": self.username,
                    "password": self.password,
                }
            else:
                self.browser_proxy = {"server": self.plain_proxy}

        self.http_proxy = (
            {"http": self.http_proxy, "https": self.http_proxy} if self.proxy else None
        )

        self.phttpx = httpx.AsyncClient(proxies={"http://": self.plain_proxy})
        self.httpx = httpx.AsyncClient()

        await self.check_proxy()

    async def __adel__(self) -> None:
        await self.httpx.aclose()
        await self.phttpx.aclose()

    def split_helper(self, split) -> None:
        if not any(_.isdigit() for _ in split):
            raise SplitError("No ProxyPort could be detected")
        if split[1].isdigit():
            self.ip, self.port, self.username, self.password = split
        elif split[3].isdigit():
            self.username, self.password, self.ip, self.port = split
        else:
            raise SplitError(f"Proxy Format ({self.proxy}) isn't supported")

    def split_proxy(self) -> None:
        split = self.proxy.split(":")
        if len(split) == 2:
            self.ip, self.port = split
        elif len(split) == 3:
            if "@" in self.proxy:
                helper = [_.split(":") for _ in self.proxy.split("@")]
                split = [x for y in helper for x in y]
                self.split_helper(split)
            else:
                raise SplitError(f"Proxy Format ({self.proxy}) isn't supported")
        elif len(split) == 4:
            self.split_helper(split)
        else:
            raise SplitError(f"Proxy Format ({self.proxy}) isn't supported")

    async def check_proxy(self) -> None:
        try:
            ip_request = await self.phttpx.get(
                "https://api.ipify.org?format=json", timeout=self.timeout
            )
            ip = ip_request.json().get("ip")
        except Exception:
            raise ProxyCheckError("Could not get IP-Address of Proxy (Proxy is Invalid/Timed Out)")
        try:
            r = await self.httpx.get(f"http://ip-api.com/json/{ip}", timeout=self.timeout)
            data = r.json()
            self.country = data.get("country")
            self.country_code = data.get("countryCode")
            self.region = data.get("regionName")
            self.city = data.get("city")
            self.zip = data.get("zip")
            self.latitude = data.get("lat")
            self.longitude = data.get("lon")
            self.timezone = data.get("timezone")
            if not self.country:
                raise ProxyCheckError(
                    "Could not get GeoInformation from proxy (Proxy is probably not Indexed)"
                )
        except Exception as e:
            raise ProxyCheckError(
                "Could not get GeoInformation from proxy (Proxy is probably not Indexed)"
            ) from e



import re
from dataclasses import dataclass
from datetime import datetime, timedelta
from http.client import responses as status_codes
from json import detect_encoding
from typing import Callable, Iterable, List, Literal, Optional, Union

from orjson import dumps, loads

import playfulrequests
from playfulrequests.cffi import PORT
from playfulrequests.exceptions import ClientException

from .cookies import RequestsCookieJar
from .toolbelt import CaseInsensitiveDict, FileUtils


class ProcessResponse:
    def __init__(
        self,
        session,
        method: str,
        url: str,
        files: Optional[dict] = None,
        cookies: Optional[Union[RequestsCookieJar, dict, list]] = None,
        **kwargs,
    ) -> None:
        self.session: 'playfulrequests.session.TLSSession' = session
        self.method: str = method
        self.url: str = url

        if files:
            data = kwargs['data']
            headers = kwargs['headers']
            # assert that data is a dict
            if data is not None:
                assert isinstance(data, dict), "Data must be a dict when files are passed"
            # convert files to multipart/form-data
            kwargs['data'], content_type = FileUtils.encode_files(files, data)
            # content_type needs to be set to Content-Type header
            if headers is None:
                headers = {}
            # else if headers were provided, append Content-Type to those
            elif isinstance(headers, dict):
                headers = CaseInsensitiveDict(headers)
            headers['Content-Type'] = content_type
            kwargs['headers'] = headers

        self.cookies: Optional[Union[RequestsCookieJar, dict, list]] = cookies
        self.kwargs: dict = kwargs
        self.response: Response

    def send(self) -> None:
        time: datetime = datetime.now()
        self.response = self.execute_request()
        self.response.elapsed = datetime.now() - time

    def execute_request(self) -> 'Response':
        try:
            resp = self.session.execute_request(
                method=self.method,
                url=self.url,
                cookies=self.cookies,
                **self.kwargs,
            )
        except IOError as e:
            raise ClientException('Connection error') from e
        resp.session = None if self.session.temp else self.session
        resp.browser: str = self.session.browser
        return resp


class ProcessResponsePool:
    '''
    Processes a pool of ProcessResponse objects
    '''

    def __init__(self, pool: List[ProcessResponse]) -> None:
        self.pool: List[ProcessResponse] = pool

    def execute_pool(self) -> List['Response']:
        values: list = []
        for proc in self.pool:
            # get the request data
            payload, headers = proc.session.build_request(
                method=proc.method,
                url=proc.url,
                cookies=proc.cookies,
                **proc.kwargs,
            )
            # remember full set of headers (including from session)
            proc.full_headers = headers
            # add to values
            values.append(payload)
        # execute the pool
        try:
            # send request
            resp = proc.session.server.post(
                f'http://127.0.0.1:{PORT}/multirequest', body=dumps(values)
            )
            response_object = loads(resp.read())
        except Exception as e:
            raise ClientException('Connection error') from e
        # process responses
        return [
            proc.session.build_response(proc.url, proc.full_headers, data)
            for proc, data in zip(self.pool, response_object)
        ]


@dataclass
class Response:
    """
    Response object

    Methods:
        json: Returns the response body as json
        render: Renders the response body with BrowserSession
        find: Shortcut to .html.find

    Attributes:
        url (str): Response url
        status_code (int): Response status code
        reason (str): Response status reason
        headers (CaseInsensitiveDict): Response headers
        cookies (RequestsCookieJar): Response cookies
        text (str): Response body as text
        content (Union[str, bytes]): Response body as bytes or str
        ok (bool): True if status code is less than 400
        elapsed (datetime.timedelta): Time elapsed between sending the request and receiving the response
        html (playfulrequest.parser.HTML): Response body as HTML parser object
    """

    url: str
    status_code: int
    headers: 'playfulrequests.client.CaseInsensitiveDict'
    cookies: RequestsCookieJar
    _text: Optional[str] = None
    _content: Optional[Union[str, bytes]] = None

    # set by ProcessResponse
    history: Optional[List['Response']] = None
    session: Optional[
        Union['playfulrequests.session.TLSSession', 'playfulrequests.browser.BrowserSession']
    ] = None
    browser: Optional[Literal['firefox', 'chrome']] = None
    elapsed: Optional[timedelta] = None

    @property
    def reason(self):
        return status_codes[self.status_code]

    def json(self, **kwargs) -> Union[dict, list]:
        # use faster json processing
        return loads(self.content, **kwargs)

    @property
    def encoding(self) -> Optional[str]:
        if type(self._content) is bytes:
            return detect_encoding(self._content)

    @property
    def content(self) -> Union[str, bytes]:  # sourcery skip: reintroduce-else
        if self._content is not None:
            return self._content
        return self._text

    @property
    def text(self) -> str:
        if self._text is not None:
            return self._text
        if type(self._content) is not bytes:
            return self._content
        self._text = self._content.decode()
        return self._text

    @property
    def html(self) -> 'playfulrequests.parser.HTML':
        if not self.__dict__.get('_html'):
            self._html = playfulrequests.parser.HTML(
                session=self.session, url=self.url, html=self.content
            )
        return self._html

    @property
    def find(self) -> Callable:
        return self.html.find
    
    @property
    def find_all(self) -> Callable:
        return self.html.find_all

    @property
    def ok(self) -> bool:
        return self.status_code < 400

    @property
    def links(self) -> dict:
        '''Returns the parsed header links of the response, if any'''
        header = self.headers.get("link")
        resolved_links = {}

        if not header:
            return resolved_links

        links = parse_header_links(header)
        for link in links:
            key = link.get("rel") or link.get("url")
            resolved_links[key] = link
        return resolved_links

    def __bool__(self) -> bool:
        '''Returns True if :attr:`status_code` is less than 400'''
        return self.ok

    def render(
        self,
        *,
        headless: bool = True,
        mock_human: bool = False,
        extensions: Optional[Union[str, Iterable[str]]] = None,
    ) -> 'playfulrequests.browser.BrowserSession':
        return playfulrequests.browser.render(
            response=self,
            session=self.session,
            proxy=self.session.proxies if self.session else None,
            headless=headless,
            mock_human=mock_human,
            extensions=extensions,
            browser=self.browser,
        )

    def __enter__(self):
        return self

    def __repr__(self):
        return f"<Response [{self.status_code}]>"


def parse_header_links(value):
    '''
    Return a list of parsed link headers proxies.
    i.e. Link: <http:/.../front.jpeg>; rel=front; type="image/jpeg",<http://.../back.jpeg>; rel=back;type="image/jpeg"
    :rtype: list
    '''
    links = []
    replace_chars = " '\""
    value = value.strip(replace_chars)

    if not value:
        return links

    for val in re.split(", *<", value):
        try:
            url, params = val.split(";", 1)
        except ValueError:
            url, params = val, ""
        link = {"url": url.strip("<> '\"")}
        for param in params.split(";"):
            try:
                key, value = param.split("=")
            except ValueError:
                break
            link[key.strip(replace_chars)] = value.strip(replace_chars)
        links.append(link)
    return links


def build_response(res: Union[dict, list], res_cookies: RequestsCookieJar) -> Response:
    '''Builds a Response object'''
    # build headers
    if res["headers"] is None:
        res_headers = {}
    else:
        res_headers = {
            header_key: header_value[0] if len(header_value) == 1 else header_value
            for header_key, header_value in res["headers"].items()
        }
    return Response(
        # add target / url
        url=res["target"],
        # add status code
        status_code=res["status"],
        # add headers
        headers=playfulrequests.client.CaseInsensitiveDict(res_headers),
        # add cookies
        cookies=res_cookies,
        # add response body
        _text=res["body"],
    )



import ctypes
import os
import socket
from platform import machine
from sys import platform
from typing import Tuple
from time import sleep

import rich.progress
from httpx import get, stream
from orjson import loads

root_dir = os.path.abspath(os.path.dirname(__file__))

# map machine architecture to playfulrequests-cgo binary name
arch_map = {
    'amd64': 'amd64',
    'x86_64': 'amd64',
    'x86': '386',
    'i686': '386',
    'i386': '386',
    'arm64': 'arm64',
    'aarch64': 'arm64',
    'armv5l': 'arm-5',
    'armv6l': 'arm-6',
    'armv7l': 'arm-7',
    'ppc64le': 'ppc64le',
    'riscv64': 'riscv64',
    's390x': 's390x',
}


class LibraryManager:
    # specify specific version of hrequests-cgo library
    BRIDGE_VERSION = '1.'

    def __init__(self):
        self.parent_path = os.path.join(root_dir, 'bin')
        self.file_cont, self.file_ext = self.get_name()
        filename = self.check_library()
        self.full_path = os.path.join(self.parent_path, filename)

    @staticmethod
    def get_name() -> Tuple[str, str]:
        try:
            arch = arch_map[machine().lower()]
        except KeyError as e:
            raise OSError('Your machine architecture is not supported.') from e
        if platform == 'darwin':
            return f'darwin-{arch}', '.dylib'
        elif platform in ('win32', 'cygwin'):
            return f'windows-4.0-{arch}', '.dll'
        return f'linux-{arch}', '.so'

    def check_library(self):
        for file in os.listdir(self.parent_path):
            if not file.endswith(self.file_ext):
                continue
            if file.startswith(f'hrequests-cgo-{self.BRIDGE_VERSION}'):
                return file
            # delete residual files from previous versions
            os.remove(os.path.join(self.parent_path, file))
        self.download_library()
        return self.check_library()

    def download_library(self):
        print('Downloading hrequests-cgo library from daijro/hrequests...')
        # pull release assets from github daijro/hrequests
        resp = get('https://api.github.com/repos/daijro/hrequests/releases/latest')
        assets = loads(resp.content)['assets']
        for asset in assets:
            if self.file_cont in asset['name'] and asset['name'].endswith(self.file_ext):
                url: str = asset['browser_download_url']
                name: str = asset['name']
                break
        else:
            raise IOError('Could not find a matching binary for your system.')
        with open(os.path.join(self.parent_path, name), 'wb') as fstream:
            self.download_file(fstream, url)

    @staticmethod
    def download_file(fstream, url):
        # file downloader with progress bar
        total: int
        with stream('GET', url, follow_redirects=True) as resp:
            total = int(resp.headers['Content-Length'])
            with rich.progress.Progress(
                    "[progress.percentage]{task.percentage:>3.0f}%",
                    rich.progress.BarColumn(bar_width=40),
                    rich.progress.DownloadColumn(),
                    rich.progress.TransferSpeedColumn(),
            ) as progress:
                download_task = progress.add_task("Download", total=total)
                for chunk in resp.iter_bytes():
                    fstream.write(chunk)
                    progress.update(download_task, completed=resp.num_bytes_downloaded)


def GetOpenPort() -> int:
    while True:
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.bind(("", 0))
            s.listen(1)
            port = s.getsockname()[1]
            s.close()
            return port
        except OSError:
            continue


class GoString(ctypes.Structure):
    _fields_ = [("p", ctypes.c_char_p), ("n", ctypes.c_longlong)]


# load the shared package
libman = LibraryManager()

library = ctypes.cdll.LoadLibrary(libman.full_path)
del libman

library.DestroySession.argtypes = [GoString]
library.DestroySession.restype = ctypes.c_void_p


def start_server(port: int):
    encoded_port = str(port).encode('utf-8')
    sleep(5)
    library.StartServer(str(port))


def destroy_session(session_id: str):
    encoded_session_id = session_id.encode('utf-8')
    library.DestroySession(GoString(encoded_session_id, len(encoded_session_id)))


# Get an open port and start the server
PORT = GetOpenPort()
print(f"Starting server on port {PORT}")
start_server(PORT)



from .element_handle import mock_element_handle
from .frame import mock_frame
from .frame_locator import mock_frame_locator
from .js_handle import mock_js_handle
from .locator import mock_locator
from .mouse import mock_mouse
from .page import new_page

from .proxy_manager import ProxyManager
from .faker import Faker
from .playwright_mock import ChromeBrowser, FirefoxBrowser



import asyncio
from os.path import dirname, join
import json
from async_class import AsyncObject, link


class Faker(AsyncObject):
    async def __ainit__(self, context, proxy, browser_name, user_agent) -> None:
        link(self, context)
        self.useragent: str = user_agent
        threads = [self.computer(proxy, browser_name), self.locale(proxy)]
        await asyncio.gather(*threads)

    async def computer(self, proxy, browser_name) -> None:
        try:
            url = f"http://fingerprints.bablosoft.com/preview?rand=0.1&tags={browser_name},Desktop,Microsoft%20Windows"
            # Sometimes the API is offline
            while True:
                r = await proxy.httpx.get(url, timeout=20)
                data = r.json()
                # self.useragent = data.get("ua")
                self.vendor: str = data.get("vendor")
                self.renderer: str = data.get("renderer")
                self.width: int = data.get("width", 1280)
                self.height: int = data.get("height", 720)
                self.avail_width: int = data.get("availWidth", 1280)
                self.avail_height: int = data.get("availHeight", 720)
        except Exception:
            # If Bablosoft Website is offline
            # self.useragent = UserAgent().msie
            self.vendor: str = "Google Inc."
            self.renderer: str = "Google Inc. (AMD)"
            self.width: int = 1280
            self.height: int = 720
            self.avail_width: int = 1280
            self.avail_height: int = 720

    async def locale(self, proxy) -> None:
        with open(join(dirname(__file__), "locale.json"), "r") as f:
            language_dict = json.load(f)

        country_code = proxy.country_code

        if country_code in language_dict:
            self.locale, self.language_code = language_dict[country_code]
        else:
            raise ValueError("Proxy Country not supported")
